/**
 * Copyright (c) 2014-2018 Enzien Audio Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "HvSignalLorenz.h"

hv_size_t sLorenz_init(SignalLorenz *o, float x, float y, float z) {
  o->xm1 = x;
  o->ym1 = y;
  o->zm1 = z;
  return 0;
}

void sLorenz_onMessage(HeavyContextInterface *_c, SignalLorenz *o, int letIndex,
    const HvMessage *const m) {
  if (msg_hasFormat(m, "fff")) {
    o->xm1 = msg_getFloat(m,0);
    o->ym1 = msg_getFloat(m,1);
    o->zm1 = msg_getFloat(m,2);
  }
}
