// Microsoft Visual C++ generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "afxres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (United States) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)

/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

{%- if parameters|length > 0 %}
STRINGTABLE
BEGIN
    {%- for k, v in parameters %}
    IDS_HV_PARAM_{{k|upper}}      "{{v.display}}"
    {%- endfor %}
END
{%- endif %}

#endif    // English (United States) resources
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
// English (Canada) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENC)
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_CAN
#pragma code_page(1252)

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE
BEGIN
    "#include ""afxres.h""\r\n"
    "\0"
END

3 TEXTINCLUDE
BEGIN
    "#define AK_WWISESDK_SAMPLE_FILEDESC ""{{name}} Wwise plug-in generated by Heavy""\r\n"
    "#define AK_WWISESDK_SAMPLE_INTERNALNAME ""{{name}}.dll""\r\n"
    "#define AK_WWISESDK_SAMPLE_ORIGINALFILE ""{{name}}.dll""\r\n"
    "#include ""AkWwiseSDKVersion.rc""\r\n"
    "#define _AFX_NO_SPLITTER_RESOURCES\r\n"
    "#define _AFX_NO_OLE_RESOURCES\r\n"
    "#define _AFX_NO_TRACKER_RESOURCES\r\n"
    "#define _AFX_NO_PROPERTY_RESOURCES\r\n"
    "\r\n"
    "#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)\r\n"
    "LANGUAGE 9, 1\r\n"
    "#pragma code_page(1252)\r\n"
    "#include ""afxres.rc""     // Standard components\r\n"
    "#endif\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_HV_{{name|upper}}_PLUGIN_BIG DIALOGEX 0, 0, 518, 280
STYLE DS_SETFONT | DS_FIXEDSYS | WS_CHILD | WS_CLIPCHILDREN
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    {%- set maxParamNameLength = 15 %}
    {%- set offset = 0 %}
    {%- if parameters|length > 0 %}
        {%- set offset = parameters|length*20+25 %}
        {%- for k, v in parameters %}
            {%- if k|length > maxParamNameLength %}
                {%- set maxParamNameLength = k|length %}
            {%- endif %}
        {%- endfor %}
    GROUPBOX        "In Parameters", IDC_STATIC_PARAMETER_IN_GROUP, 5, 3, {{maxParamNameLength*10+65}}, {{parameters|length*20+20}}
        {%- for k, v in parameters %}
    LTEXT           "{{v.display}}", IDC_STATIC_HV_PARAM_{{k|upper}}, 25, {{((loop.index-1)*20)+20}}, {{maxParamNameLength*5}}, 13, WS_TABSTOP
    LTEXT           "Class=SuperRange;Prop={{k}}", IDC_RANGE_HV_PARAM_{{k|upper}}, {{maxParamNameLength*5}}+45, {{((loop.index-1)*20)+20}}, 80, 13, WS_BORDER | WS_TABSTOP
        {%- endfor %}
    {%- endif %}
    {%- if sends|length > 0 %}
    GROUPBOX        "Out Parameters", IDC_STATIC_PARAMETER_OUT_GROUP, 5, {{offset+5}}, {{maxParamNameLength*10+65}}, {{sends|length*20+20}}
        {%- for k, v in sends %}
    LTEXT           "{{k|lower}}", IDC_STATIC_HV_PARAM_OUT_{{k|upper}}, 25, {{((loop.index-1)*20)+offset+20}}, {{maxParamNameLength*5}}, 13, WS_TABSTOP
        {%- endfor %}
        {%- set offset = offset + sends|length*20+25 %}
    {%- endif %}
    {%- if tables|length > 0 %}
        {%- for k, v in tables %}
            {%- if k|length > maxParamNameLength %}
                {%- set maxParamNameLength = k|length %}
            {%- endif %}
        {%- endfor %}
    GROUPBOX        "Tables", IDC_STATIC_TABLE_GROUP, 5, {{offset+5}}, {{maxParamNameLength*10+65}}, {{tables|length*20+20}}
        {%- for k, v in tables %}
    LTEXT           "{{v.display}}", IDC_STATIC_HV_TABLE_{{k|upper}}, 25, {{((loop.index-1)*20)+offset+20}}, {{maxParamNameLength*5}}, 13, WS_TABSTOP
    PUSHBUTTON      "Load Sample", IDC_BUTTON_HV_TABLE_{{k|upper}}, {{maxParamNameLength*5}}+45, {{((loop.index-1)*20)+offset+20}}, 50, 13
        {%- endfor %}
    {%- endif %}
END

IDD_HV_{{name|upper}}_PLUGIN_SMALL DIALOGEX 0, 0, {{(parameters|length) * 50}}, 13
STYLE DS_SETFONT | DS_FIXEDSYS | DS_CONTROL | WS_CHILD
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    {%- for k, v in parameters %}
        {% if loop.index < 4 -%}
    LTEXT           "Class=SuperRange;Prop={{k}};Title={{v.display}};", IDC_RANGE_HV_PARAM_{{k|upper}}, {{((loop.index-1)*50)}}, 0, 50, 13, WS_BORDER | WS_TABSTOP
        {%- endif %}
    {%- endfor %}
END

/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO
BEGIN
    IDD_HV_{{name|upper}}_PLUGIN_BIG, DIALOG
    BEGIN
    END

    IDD_HV_{{name|upper}}_PLUGIN_SMALL, DIALOG
    BEGIN
        RIGHTMARGIN, 196
        VERTGUIDE, 2
    END
END
#endif    // APSTUDIO_INVOKED

#endif    // English (Canada) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//
#define AK_WWISESDK_SAMPLE_FILEDESC "{{name}} Wwise plug-in generated by Heavy"
#define AK_WWISESDK_SAMPLE_INTERNALNAME "{{name}}.dll"
#define AK_WWISESDK_SAMPLE_ORIGINALFILE "{{name}}.dll"
#include "AkWwiseSDKVersion.rc"
#define _AFX_NO_SPLITTER_RESOURCES
#define _AFX_NO_OLE_RESOURCES
#define _AFX_NO_TRACKER_RESOURCES
#define _AFX_NO_PROPERTY_RESOURCES

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
LANGUAGE 9, 1
#pragma code_page(1252)
#include "afxres.rc"     // Standard components
#endif

/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED
{# force new line #}
